
# Amateur Xperimental Unconnected module
# $Id: axu.pm,v 1.3 2005/04/27 16:16:55 wa7nwp Exp $
# $Source: /cvsroot/libaxu/libaxu/axu.pm,v $

# PUBLIC ROUTINES in this module
#
# axu_position
# ($lat,$lng,$symbol,$phg,$day,$time,$text)=axu_decode_time_position($data);
# ($range, $bearing) = &axu_range_and_bearing( $latitude1, $longitude1, $latitude2, $longitude2);

# Note.  Speed is stored and transmitted on the air as Knots.
# To convert speed from KNOTS to Miles
#$speed *= 1.15077944 if $speed > 0;

require 5;

my $debugv = 1;
$debugv="";


# display the main fields of a packet.  Used for diagnostics
# Nothing returned
sub axu_displaypacket {
  my ($packet) = @_;
  my ($source,$type,$dest,$path,$data) = &axu_decode($packet);
  printf "source: $source  dest: $dest  path: $path\n";
  printf "type: %s ($type) data:   $data\n",&axu_type_name($type);
  if ( $type eq ":" ) { # message
    my ($to, $msg, $id) = &axu_decode_message($data);
    print "Message: (to) $to  (id) $id  (msg) $msg\n";
  }
  # MIC-E values - for debugging
  if (&axu_is_any_position($type)) {
    #A position packet # data to be returned
    my ($lat,$lng,$symbol,$course, $speed, $altitude) = 
        &axu_decode_any_position ($type,$dest,$data);
    printf "IsPosition: %8.4f %8.4f Symbol: $symbol Course $course Speed $speed Altitude $altitude \n", $lat, $lng;
  }
}

# Is this packet any type of a position packet?
# my ($IsPosition) = axu_is_any_position( $type );
sub axu_is_any_position {
   my ($type)=@_;
   my $IsPosition = 0;
   $IsPosition = 1 if ( $type =~ /[\`\'\@\!\;\=\$]/ );
   return $IsPosition;
}

# Decode basic data from any  position packet?
# Calling:
#    my ($lat,$lng,$symbol,$course, $speed, $altitude) = 
#        &axu_decode_any_position ($type,$dest,$data);

sub axu_decode_any_position {
    my ($type,$dest,$data) = @_;
    # data to be returned
    my ($lat,$lng,$symbol,$course, $speed, $altitude);
    # scratch
    my ($valid, $status, $phg, $text);
    if ( $type =~ m/\@|\!|\=/ ) {
      ($lat,$lng,$symbol,$phg,$course,$speed,$altitude,$text)=axu_decode_position($data);
    } elsif ( $type =~ m/\;/ ) {
      ($lat,$lng,$symbol,$phg,$day,$time,$text)=axu_decode_time_position($data);
    } elsif ( $type =~ m/[\`\']/ ) {
      ($valid, $lat,$lng,$symbol,$status,$course,$speed,$altitude,$text)=
         axu_decode_MIC_E($dest,$data);
    } elsif ( $type eq "\$" ) {
      ($lat,$lng,$gpstype,$isvalid,$bearing,$speed,$altitude)=axu_decode_gps($data);
      $symbol = &axu_symbol_gps( $dest );
      # if it was a short GPS, not GPSxx, entry, use the SSID
      # No source callsign available in this function
      #$symbol = &axu_symbol_ssid( $source ) if ($symbol =~ /^\?/);
    }
   return ($lat,$lng,$symbol,$course, $speed, $altitude);
}


# Gated 3rd party '}'
# Mic-E current '''
# Mic-E old data '`'
# Posit computerized '='
# Posit non-APRS '!'
# Status '>'
# Weather '_'   (UL is special case)
#         !! is Peet II
#         '#'
# Mobiles '$'
#  -  $UL is a weather packet
# Telemetry 'T'
# Object ';'
# Object - deleted "_"
#
sub axu_type_name{
  my ($type) = @_;
  my $type_name = "Unknown";
  $type_name = "3rdParty" if $type eq "}";
  $type_name = "position@" if $type eq "\@";
  $type_name = "position\$" if $type eq "\$";
  $type_name = "MicE" if $type eq "'";
  $type_name = "MicEold" if $type eq "`";
  $type_name = "position=" if $type eq "=";
  $type_name = "position" if $type eq "!";
  $type_name = "status" if $type eq ">";
  $type_name = "weather" if $type eq "_";
  $type_name = "message" if $type eq ":";
  $type_name = "telemetry" if $type eq "T";
  $type_name = "object" if $type eq ";";
  $type_name = "ObjectDeleted" if $type eq ";";
  return $type_name;
}



sub axu_datafile  {
    return "/tmp/axudata.txt" ;
}

# Send a packet  -  use configuration to see how to send it.
# options:  AX25 beacon, (soon: UDP, TCP)
sub axu_send {
  my($call,$dest,$path,$port,$msg)=@_;
  &axu_send_beacon($call,$dest,$path,$port,$msg);
}

# Send a packet using AX25 Beacon command
sub axu_send_beacon {
  my($call,$dest,$path,$port,$msg)=@_;

  $path =~ s/,/ /g;  # need spaces in path, not commas
  my $fullpath;

  # "%" is % to signify no VIA's
  if ( $path eq "%" ) {
     $fullpath = sprintf("%s",$dest);  # no path
  } else {
     $fullpath = sprintf("%s via %s",$dest,$path);
  }

  my $outstring= "/usr/sbin/beacon";
  $outstring .= " -c " . $call . " -s -d \'". $fullpath . "\' ";
  $outstring .= $port . " \'" . $msg . "\'";
printf " system $outstring\n";
  system $outstring;

  #system "/usr/sbin/beacon -c wa7nwp-14 -s -d \'". $fullpath . "\' 439 \'" . $fullmsg . "\'";
}



# break a packet into core pieces

# my($source,$type,$dest,$path,$data) = &axu(packet)
sub axu_decode    {
    my($arg)=@_;
    my ($source,$rest,$prepath,$data,$dest,$path);
    ($source,$rest) = split(/>/, $arg, 2);
    ($prepath,$data) = split(/:/, $rest, 2);
    if ( $prepath =~ />/ ) {
        ($path,$dest) = split(/>/, $prepath, 2);
    } else {
        ($dest,$path) = split(/,/, $prepath, 2);
    }
    $data =~ s/[
 ]*$//;   # remove possible CR and space at end
    $type = "?";
    if ($dest =~ /^gp|ap|beacon/i ) {
        $type = substr( $data, 0, 1 );
        if ($type =~ /[\:\'\`\=\!\#\>\;\_\$\}\@]/) {
            $data =~ /^\$ULT/ &&  ($type = "_") ; # Ultimeter 100 $ULT format
            $data =~ /^!!/ &&  ($type = "_") ;   # Ultimeter 100 !! format
            $data =~ /RSW$/ &&  ($type = "_") ;  # RS W with position
            $type =~ /^\#/ &&  ($type = "_") ;  # Peet U-II #7006700000000
        } else {
            $type = "?";
            $data =~ /^T#/ &&   ($type = "T") ;
            $data =~ /\:ack.*$/i &&  ($type = ":") ;
            $data =~ /[\w ]+\:.*\{.+$/ &&  ($type = ":") ;
        }
    } else {
      # might be MIC-E
      if ($data =~ /^([\'\`])/ ) {
        $type = $1;
      }
    }
    return $source,$type,$dest,$path,$data;
    }

# Gated 3rd party '}'
# Mic-E current '''
# Mic-E old data '`'
# Posit computerized '='
# Posit non-APRS '!'
# Status '>'
# Weather '_'   (UL is special case)
#         !! is Peet II
#         '#'
# Mobiles '$'
#  -  $UL is a weather packet
# Telemetry 'T'
# Object ';'
# Object - deleted "_"


# ###  Decode a Message

# source>   ":AC7KP-5  :Will do.{13}L"
# usage:  my ($to_call, $txt, $msg_id) = &axu_decode_message( $data );

sub axu_decode_message
{
   my ($data) = @_;
   
   # break up data in to portions of a message
   $data =~ /:{0,1}(\S+)\s*?:([^{]+)(.*)/;
     ($to_call, $txt, $msg_id) = ( $1, $2, $3);

   # Some packet carry an ACK, the "}L" in the example.
   # Strip it if it's in the message_id
   $msg_id =~ s/}.*$//;

   # If this is a BLN message, use that for the ID
   if ( $to_call =~ /^(BLN\w*)/i ) {
      $msg_id = "$1";
print "MSG BLN:  message_id = $msg_id\n" if $debugx;
   }

   # if msg_id is still blank, look for a period terminated
   # short word at the start of the message.  This covers
   # special messages like  EQN., PARM. and UNITS.
   unless ($msg_id) {
      if ( $txt =~ /^(\w{1,10})\./ ) {
          $msg_id = $1;
      }
   }

print "sender:  $sender_id   To: $to_call  Msg ID: $msg_id  Message: $txt \n" if $debugx;

  return ($to_call, $txt, $msg_id);
}

### old

# my($from,$msg,$ack1,$ack2) = &axu_msgdecode($data);
# WA7NWP-14 :blah blah {13}
#sub axu_msgdecode    {
#    my($arg)=@_;
#    my($from,$msg,$ack1,$ack2);
#    ($from,$rest) = split(/:/, $arg, 2);
#    ($msg,$rest) = split(/{/, $rest, 2);
#    ($ack1,$ack2) = split(/\}/, $rest, 2);
#    return $from,$msg,$ack1,$ack2;
#}


# ###  Decode an object
#
# pos file entry
# 53.87 !4826.41N/12205.40WrPHG5550 52.17/53.87 T100 W7SIX 43825 Grass Mountain
sub axu_obj_decode {
    my ($arg)=@_;
    my ( $obj_name, $obj_lat, $obj_lng, $obj_sym, $obj_text);
    $arg =~ /(^.*)!(........)(.)(.........)(.)(.*)/;
#print "1 $1 2 $2 3 $3 4 $4 5 $5 6 $6\n";
    $obj_name = $1;
    $obj_lat = axu_decode_lat($2);
    $obj_lng = axu_decode_long($4);
    $obj_sym  = $3 . $5;
    $obj_text = $6;
    return $obj_name, $obj_lat, $obj_lng, $obj_sym, $obj_text;
}


# create position string from Lat and Long
sub axu_encode_position {
    my($lat,$long,$symbol,$phg)=@_;
    my($posit);
    my($szlat) = axu_encode_lat($lat);
    my($szlong)= axu_encode_long($long);
    my($sym1,$sym2)=(substr($symbol,0,1),substr($symbol,1,1));
    $posit = sprintf("%8s%s%9s%s",
        $szlat,$sym1,$szlong,$sym2);
    return($posit);
}

### decode Lat and Long from APRS position string

# Invocation
#  ($lat,$lng,$symbol,$phg,$course,$speed,$altitude,$text)=axu_decode_position($data);
sub axu_decode_position {
   my($posit)=@_;
   my($lat, $long, $symbol, $phg, $text, $course, $speed, $altitude) =
      (0.0, 0.0, "//", "0000", "", 0.0, 0.0, 0.0);
   if ( ($posit =~ /.(\d\d[\d ][\d ].[\d ]{2}\w)(.)(\d\d\d[\d ][\d ].[\d ]{2}\w)(.)(.*)$/)) {
        $lat = axu_decode_lat($1);
        $long = axu_decode_long($3);
        $symbol = axu_symbol_overlay($2 . $4);
        $text = $5;
        $phg = "";
        # extract PHG entry from text field
	if ($text =~ /PHG(\d{4})(.*)/) {
          $phg = $1;
          $text = $2;
        }
        # extract Altitude from text field
        if ( $text =~ m/A\=(\d+)/ ) {
          $altitude = $1;
          $text =~ s/A\=\d+//;
        }
        # extract course and speed if we have a  NNN/NNN at the head of the line
        if ( $text =~ m/^(\d\d\d)\/(\d\d\d)\// ) {
          $course = $1;
          $speed = $2;
          $text =~ s/^(\d\d\d)\/(\d\d\d)\///;
        }
        # if first character of the remaining text is a /, remove it
        $text =~ s/^\///;
    }
#else { printf "NO match:\n"; }
    return($lat,$long,$symbol,$phg,$course,$speed,$altitude,$text);
}


# Invocation
#  ($lat,$lng,$symbol,$phg,$course,$speed,$altitude,$text)=axu_decode_time_position($data);
sub axu_decode_time_position {
   my($posit)=@_;
   my($lat, $long, $symbol, $phg, $text, $course, $speed, $altitude, $time) =
      (0.0, 0.0, "//", "0000", "", 0.0, 0.0, 0.0, "");
   if ( ($posit =~ /.(\d{6})z(\d\d[\d ][\d ].[\d ]{2}\w)(.)(\d\d\d[\d ][\d ].[\d ]{2}\w)(.)(.*)$/)) {
        $time = $1;
        $lat = axu_decode_lat($2);
        $long = axu_decode_long($4);
        $symbol = axu_symbol_overlay($3 . $5);
        $text = $6;
        $phg = "";
        # extract PHG entry from text field
	if ($text =~ /PHG(\d{4})(.*)/) {
          $phg = $1;
          $text = $2;
        }
        # extract Altitude from text field
        if ( $text =~ m/A\=(\d+)/ ) {
          $altitude = $1;
          $text =~ s/A\=\d+//;
        }
        # extract course and speed if we have a  NNN/NNN at the head of the line
        if ( $text =~ m/^(\d\d\d)\/(\d\d\d)\// ) {
          $course = $1;
          $speed = $2;
          $text =~ s/^(\d\d\d)\/(\d\d\d)\///;
        }
        # if first character of the remaining text is a /, remove it
        $text =~ s/^\///;
    }
#else { printf "NO match:\n"; }
    return($lat,$long,$symbol,$phg,$course,$speed,$altitude,$time . text);
}

# decode Lat and Long from GPS generated position string
# GPRMC string
# $GPRMC,014423,A,4744.3150,N,12221.2342,W,000.0,292.0,011000,018.9,E*6C
#   Recommended Minimum Specific GPS/Transit Data (RMC)
#   $GPRMC,hhmmss.ss,a,ddmm.mmmm,N,dddmm.mmmm,W,z.z,y.y,ddmmyy,d.d*CC
#   hhmmss.ss - time
#   a - status (A=Valid, V = Invalid)
#   ddmm.mmmm,N - latitude
#   dddmm.mmmm,W - longitude
#   z.z - speed over ground
#   y.y - heading, true north
#   ddmmyy - UTC time of fix
#   d.d - magnetic variation
#   v - variation sense
#   CC - checksum
# $GPGGA,221326,4736.8945,N,12219.4070,W,1,08,1.3,143.2,M,-18.4,M,,*74 
#   GPS Fix Data (GGA)
#   $GPGGA,hhmmss.ss,ddmm.mmmm,N,dddmm.mmmm,W,q,ss,y.y,a.a,M,g.g,M,t.t,iiii*CC
#   hhmmss.ss - UTC time of fix
#   ddmm.mmmm,N - latitude
#   dddmm.mmmm,W - longitude
#   q - GPS quality (0 -gps not available, 1-gps available, 2 -differential fix available)
#   ss - number of satellites being used (must be at least four fora three dimensional fix)
#   y.y - DHOP (dilution of horizontal position)
#   a.a,M - antenna height in meters
#   g.g,M - geoidal separation in meters
#   t.t - age of differential data
#   iiii - differential reference station ID
#   CC - checksum
#
# my($lat,$long,$gpstype,$isvalidgps,$course,$speed,$altitude) = axu_decode_gps($data);
#
sub axu_decode_gps {
   my($gps_posit)=@_;
   my($lat,$long,$speed,$course,$altitude) = ( 0.0, 0.0, 0.0, 0.0, 0.0 );
   #$lat=0.0 ; $long = 0.0;
   my($gpstype,$remaining) = split(/,/, $gps_posit, 2);
   my($gpstime,$isvalidgps,$gpslat,$gpslatX,$gpslng,$gpslngX);
   my($x1,$x2,$x3,$x4,$x5) ;
  # GPRMC
   ($gpstime,$isvalidgps,$gpslat,$gpslatX,$gpslng,$gpslngX,$speed,$course,$x3,$x4,$x5) =
      split(/,/, $remaining, 11) if ($gpstype eq "\$GPRMC");
  # GPGGA
   ($gpstime,$gpslat,$gpslatX,$gpslng,$gpslngX,$x1,$x2,$x3,$altitude,$x5) =
      split(/,/, $remaining, 11) if ($gpstype eq "\$GPGGA");
   $lat = axu_decode_lat($gpslat . $gpslatX);
   $long = axu_decode_long($gpslng . $gpslngX);

   # adjust meters altitude to feet
   $altitude *= 3.2808 if ($altitude != 0);

   return($lat,$long,$gpstype,$isvalidgps,$course,$speed,$altitude);
}


# decode MIC-E string

sub axu_decode_MIC_E {
   my($tofield,$datafield)=@_;
   my($elat,$symbol,$elng,$status,$course,$speed,$altitude,$message);
   $symbol="//";
   $status=0;
   $course=0;
   $speed=0;
   $altitude=0;
   $message = "";
   $message = substr($datafield,9) if (length($datafield) > 9);

   # Mic E fields
   $elat = 0.0;
   $elng = 0.0;
   $emsg = 0;
   $eLatHem = "S";
   $eLonHem = "E";
   $eLonOffset = 0;

   # strip SSID from ToField if it is there
   $tofield =~ s/\-\d{1,2}//;

   return(0) if length($tofield) != 6;
   # process the 6 characters of the information field

     ### char 1 of 6
     $cval = ord(substr($tofield,0,1))-48;
     if ( $cval > 16 ) { $cval -= 32; $status = 4;}
     $elat += 10*$cval;

     ### char 2 of 6
     $cval = ord(substr($tofield,1,1))-48;
     if ( $cval > 16 ) { $cval -= 32; $status += 2; }
     $elat += $cval;

     my $minutes = 0.0;
     ### char 3 of 6
     $cval = ord(substr($tofield,2,1))-48;
     if ( $cval > 16 ) { $cval -= 32; $status += 1;}
     $minutes = $cval * 10;

     ### char 4 of 6
     $cval = ord(substr($tofield,3,1))-48;
     if ( $cval > 16 ) { $cval -= 32; $eLatHem = "N"; }
     $minutes += $cval;

     ### char 5 of 6
     $cval = ord(substr($tofield,4,1))-48;
     if ( $cval > 16 ) { $cval -= 32; $eLonOffset = 100;}
     $minutes += .1 * $cval;

     ### char 6 of 6
     $cval = ord(substr($tofield,5,1))-48;
     if ( $cval > 16 ) { $cval -= 32; $eLonHem = "W"; }
     $minutes += .01 * $cval;

     # convert minutes to degrees and add to total
     $elat += $minutes/60.0;

     # make negative if it is southern hemisphere
     $elat = - $elat if $eLatHem eq "S";

   ### Now to work on the longitude and other data from the true data

     # first character is type indicator.  ' or `
     # Note D700 uses ', instead of ` for current GPS data
     # - if byte 1 is ' and byte 10 is ], then it is D700

     ### char 2 of 9
     #   this uses the longitude offset value from to field
     $cval = ord(substr($datafield,1,1))-28;
     $elon = $cval + $eLonOffset;
     #   If between 180 and 189, then really 100 to 109
     #   If between 190 and 199, then really 0 to 9
     if ( $elon >= 180 && $elon <=189 )
        { $elon -= 80; }
     elsif ( $elon >=190 && $elon <= 199 )
        { $elon -= 190; }

     ### char 3 of 9 - longitude minutes
     $cval = ord(substr($datafield,2,1))-28;
     #  If over 60 minutes, modulo 60 
     $minutes = $cval >= 60 ? $cval - 60 : $cval;

     ### char 4 of 9 - longitude fraction - add to minutes value
     $cval = ord(substr($datafield,3,1))-28;
     $minutes += $cval / 100.0;

     # now we have minutes, add it to Longitude
     $elon += $minutes / 60.0;

     # negative if longitude is west
     $elon = - $elon if ( $eLonHem eq "W" );

     ### char 5 of 9 - speed in knots  tens and hundreds
     $cval = ord(substr($datafield,4,1))-28;
     $speed = $cval * 10;
     
     ### char 6 of 9 - units of speed and course in hundreds of degrees
     #   value is SSC  SS for speed and C for course hundreds
     $cval = ord(substr($datafield,5,1))-28;
     $speed += int ($cval/10);
     $speed -= 800 if $speed >= 800;
     $course = 100 * ($cval - 10 * ( int ($cval / 10 ) ));

     ### char 7 of 9 - tens and units of degrees
     $cval = ord(substr($datafield,6,1))-28;
     $course += $cval;
     $course -= 400 if ($course >= 400);

    ### char 9 of 9 is table, char 8 is symbol
    $symbol = substr($datafield,8,1) . substr($datafield,7,1);

   ### char 10 is ] for TM-D700, > for TH-D7

   #   Telemetry data 

   ### char 10 is ` for 2 printable hex values, channels 1 and 3
   ### char 10 is ' for 5 printable hex values
   ### char 10 is 0x1d then 5 binary values follow

   # Maindenhead Locator  /G is symbol.  (spc) text comment follow
   #  IO91SX/G_Hello
   #  >IO91SX/G_Hello  from TH-D7
   #  ]IO91SX/G_Hello from D700

   #  xxx} for altitude where xxx is meaters relative to 10km below sea level

   # remove D7 or D700 flags from message
   $message =~ s/^([\>\]])//;
   #  $kenwood = $1;

   # remove altitude from message 
   # Altitude is meters relative to 10,000  base 91
   $altitude = 0;
   $message =~ s/^(...)\}//;

   # This should work just testing on $1, but if it does not match, then
   # $1 has the contents of the previous match and is still non empty.
   if ( length($1) == 3 ) {
      # first byte - 91^2
      my $c = ord(substr($1,0,1))-33;
      $altitude = $c*91*91;

      # second byte - 91
         $c = ord(substr($1,1,1))-33;
      $altitude += $c*91;

      # final byte - 91
         $c = ord(substr($1,2,1))-33;
      $altitude += $c;

      # subtract 10000 bias
      $altitude -= 10000;

      # convert to feet
      $altitude *= 3.2808;
   
   }

   return(1, $elat, $elon, $symbol, $status, $course, $speed, $altitude, $message);
}


sub axu_mic_e_status {
#' MIC-E compressed packet functions
#   'assign strings to MicE message codes
   my ($intMsgCode) = @_;
   my $aprsMsgCode = "unknown";
   $aprsMsgCode = "Off Duty" if $intMsgCode == 7;
   $aprsMsgCode = "Enroute" if $intMsgCode == 6;
   $aprsMsgCode = "In Service" if $intMsgCode == 5;
   $aprsMsgCode = "Returning" if $intMsgCode == 4;
   $aprsMsgCode = "Committed" if $intMsgCode == 3;
   $aprsMsgCode = "Special" if $intMsgCode == 2;
   $aprsMsgCode = "Priority" if $intMsgCode == 1;
   $aprsMsgCode = "Emergency" if $intMsgCode == 0;
   return $aprsMsgCode;
}


# decode Lat and Long from MIC-E converted string
# 
# @082139z4536.94N/12237.47Wv317/000/E>dsy/M0/Off duty.. ]"3d} 
# @082140z4746.32N/12211.24Wj286/000/E>dsy/M2/In Service ]"4+}147.200+ PL-123hz Call me 
# @082141z4822.87N/12230.66W-004/000/E>dsy/M2/In Service ]George in LaConner 
# @082142z4733.93N/12237.73Wk015/000/E>dsy/M2/In Service
# @082042z4607.76N/12256.11W>237/000/pocketAPRS v1.08 
#
# Or could be weather string
# AA7OA>APS200,KOPEAK,K7NWS-10*:@130323z4610.86N/12350.90W_315/000g000t061r000p001P001b10114h91
#
sub axu_decode_at {
   my($posit)=@_;
   $posit =~ s/^\@//;  # remove @ if it's at the head of the line
   my $valid = 0;
   my($datetime,$lat,$sym1,$lng,$sym2,$course,$speed,$remaining) =  
      $posit =~ /(.......)(........)(.)(.........)(.)(...)\/(...)\/(.*)/;
   # a cheap test to see if the above assignment worked
   if ( "$sym2" ) {
      $valid = 1;
      $lat = axu_decode_lat($lat);
      $lng = axu_decode_long($lng);
   }
   # if it didn't, it's probably a weather string

   return($valid, $lat, $lng, $sym1 . $sym2, $datetime, $course, $speed, $remaining);
}

# Or could be weather string
# AA7OA>APS200,KOPEAK,K7NWS-10*:@130323z4610.86N/12350.90W_315/000g000t061r000p001P001b10114h91
#
sub axu_decode_at_wx {
   my($posit)=@_;
   $posit =~ s/^\@//;  # remove @ if it's at the head of the line
   my $valid = 0;
   my($datetime,$lat,$symbol,$lng,$phg,$course,$speed,$altitude,$text);
   ($lat,$lng,$symbol,$phg,$text)=axu_decode_position($data);
   # a cheap test to see if the above assignment worked
   return($valid, $lat, $lng, $symbol, $datetime, $remaining);
}

# ###  Handle the various format of weather strings

# decode Wx string
sub axu_decode_wx {
   my($wx)=@_;
   my($temp, $direction, $wind, $precip, $text, $pressure, $humidity) =
     ( 0.0,           0,     0,       0, "",            0,         0);
   if ( $wx =~ /^!/ )
   { #  UII
# Header = !! 
# Data Fields 
# 1. Wind Speed (0.1 kph) 
# 2. Wind Direction (0-255) 
# 3. Outdoor Temp (0.1 deg F) 
# 4. Rain* Long Term Total (0.01 inches) 
# 5. Barometer (0.1 mbar) 
# 6. Indoor Temp (0.1 deg F) 
# 7. Outdoor Humidity (0.1%) 
# 8. Indoor Humidity (0.1%) 
# 9. Date (day of year) 
# 10. Time (minute of day) 
# 11. Today's Rain Total (0.01 inches)* 
# not decoded
# 12. 1 Minute Wind Speed Average (0.1kph)* 
#
# ww7ra: !!0000001001810C2C27B502AA--------001304EF
#
      my(       $zWind,$z2,   $z3,  $z4,  $z5,  $z6,  $z7,  $z8,  $z9, $z10, $z11, $z ) =
        $wx =~ /!{1,2}(....)(....)(....)(....)(....)(....)(....)(....)(....)(....)(.*)/ ;
# ww7ra:           !!  0000  0010  0181  0C2C  27B5  02AA  ----  ----  0013  04EF

      $wind=hex($zWind)/10;
      $direction=hex($z2);
      $temp=hex($z3)/10;
      $pressure=hex($z5)/10 if ( $z5 =~ /\d/ );
      $humidity=hex($z7)/10;
      $precip=hex($z11)/100;
      $text="UII " . $z;
   }
   elsif ( $wx =~ /^\$ULT/ ) {
     # From: http://www.peetbros.com/HTML_Pages/faqs.htm#Q5
	#Header = $ULTW 
	#Data Fields 
	#1. Wind Speed Peak over last 5 min. (0.1 kph) 
	#2. Wind Direction of Wind Speed Peak (0-255) 
	#3. Current Outdoor Temp (0.1 deg F) 
	#4. Rain Long Term Total (0.01 in.) 
	#5. Current Barometer (0.1 mbar) 
	#6. Barometer Delta Value(0.1 mbar) 
	#7. Barometer Corr. Factor(LSW) 
	#8. Barometer Corr. Factor(MSW) 
	#9. Current Outdoor Humidity (0.1%) 
	#10. Date (day of year) 
	#11. Time (minute of day) 
	#12. Today's Rain Total (0.01 inches)* 
	#13. 1 Minute Wind Speed Average (0.1kph)* 
	#Carriage Return & Line Feed 
	# N7RIG-6 $ULTW00A9006C01C00000----000086A00001----0012026700000096
	# W7KKE-3 $ULTW007900EB020101E027E2FFFB86D1000103B0001203ED0000004A

      my(              $z1,  $z2,  $z3,  $z4,  $z5,  $z6,  $z7,  $z8,  $z9,  $z10, $z11, $z12, $z ) =
        $wx =~ /\$ULTW(....)(....)(....)(....)(....)(....)(....)(....)(....)(....)(....)(....)(.*)/ ;
#                      00A9  006C  01C0  0000  ----  0000  86A0  0001  ----  0012  0267  0000 0096
      $temp=hex($z3)/10;
      $wind=hex($z1)/10;
      $direction=hex($z2);
      $pressure=hex($z5)/10 if ( $z5 =~ /\d/ );
      $humidity=hex($z9)/10 if ( $z9 =~ /\d/ );
      $precip=hex($z12)/100;
      $text="UII " . $z;
   }
   elsif ( ($wx =~ /^\#/) or (length($wx) eq 13) )
   { # assuming it's a II-W weather packet
# BH: byte 1 == wind direction
# byte 2,3 == wind speed
# byte 4,5 == temperature, offset by -56 degrees
# byte 6,7,8,9 == rain fall
# bytes, 10,11,12,13 == pressure

#:case 'UII'  UII
#    #7006700000000
#    #E006600000000
#    wdirection(16 Base larg[1])o'N' 'NNE' 'NE' 'ENE' 'E' 'ESE' 'SE' 'SSE' 'S' 'SSW' 'SW' 'WSW' 'W' 'WNW' 'NW' 'NNW' '??'
#    wspeed16 Base larg[2 3]
#    :if '#'=0olarg  wspeed1.15xwspeed  :end
#    wspeed'I3' fmt wspeed
#    wt'M<->I3' fmt 56+16 Base larg[4 5]
#    wr'I4' fmt 16 Base larg[6 7 8 9]
#    wp'I4' fmt 16 Base larg[10 11 12 13]

      # remove "#" if present
      $wx =~ s/^\#//;

      $direction=hex(substr($wx,0,1))*22.5;
      $wind=hex(substr($wx,1,2));
      $temp=hex(substr($wx,3,2))-56;
      $precip=hex(substr($wx,5,4));
      $pressure=hex(substr($wx,9,4));
      $text="II-2 " . $wx;
   }
   elsif  ($wx =~ /^\@/ )
   { # if @date-pos-blah-blah-RSW, filter out the position
     # @230017z4744.13N/12212.05W_105/000g000t048r000P000p000h97b09960XRSW
      #c = wind direction (in degrees).
      #s = sustained one-minute wind speed (in mph).
      #g = gust (peak wind speed in mph in the last 5 minutes).
      #t = temperature (in degrees Fahrenheit). Temperatures below
      #zero are expressed as -01 to -99.
      #r = rainfall (in hundredths of an inch) in the last hour.
      #p = rainfall (in hundredths of an inch) in the last 24 hours.
      #P = rainfall (in hundredths of an inch) since midnight.
      #h = humidity (in %. 00 = 100%).
      #b = barometric pressure (in tenths of millibars/tenths of hPascal).
     #
     my(       $zdatetime,      $zc,   $zs, $zg, $zt, $zr, $zP, $zp, $zh, $zb, $zw ) =
      $wx =~ /[@]{0,1}(.*)z.{19}(...)\/(.*)g(.*)t(.*)r(.*)P(.*)p(.*)h(.*)b(.....)(.*)/ ;
     #         @    230017z      105  /000 g000 t048 r000 P000 p000 h97  b09960 XRSW
     $temp=$zt;
     $wind=$zs;
     # $zg = Gust
     $direction=$zc;
     $precip= $zp ? $zp/100 : 0;
     $pressure=$zb ? $zb/10 : 0;
     $humidity=$zh ? $zh/1 : 0;
     $text="RSW " . $zw;
   }
   else
   { # DAVIS34 or RS
     # remove "_" if present
     #$wx =~ s/^_//;
     # date  
     #        12251827c130s000g000t043r000p016P016h99b10276wU2K
     # NS7C: _01202026c132s000g000t039r000p000P000h95b10199tRSW

     #           12251827 c130 s000 g000 t043 r000 p016 P016  h99 b10276 wU2K
     my(       $zdatetime, $zc, $zs, $zg, $zt, $zr, $zp, $zP, $zh, $zb, $zw ) =
      $wx =~ /[_]{0,1}(.*)c(.*)s(.*)g(.*)t(.*)r(.*)p(.*)P(.*)h(.*)b(.*)[wt](.*)/ ;
     $temp=$zt;
     $wind=$zs;
     # $zg = Gust
     $direction=$zc;
     $precip= $zp ? $zp/100 : 0;
     $pressure=$zb ? $zb/10 : 0;
     $humidity=$zh ? $zh/1 : 0;
     $text="Davis " . $zw;
   }
   return($temp,$wind,$direction,$precip,$pressure,$humidity,$text);
}


# ### decode Telemetry string

# Examples:
# T#822,031,118,119,147,112,00011000
# T#066,123,041,002,001,000,00000000,AWT TRK 1.23
#
sub axu_decode_telemetry {
   my($data)=@_;
   # May or may not start with the T# - remove it if it's there
   $data =~ s/^T\#//;
   my( $count, $val1, $val2, $val3, $val4, $val5, $bin, $txt ) =
       split(/,/,$data,8)  ;
   return( $count, $val1, $val2, $val3, $val4, $val5, $bin, $txt );
}

# convert text posit (DDDMM.pp[NWES]) to degrees (+/-)DDD.pp
#  West and South are -
sub axu_encode_posit{
   my($posit)=@_;
}

# decode degrees posit (DDDMM.pp[NWES]) to (+/-)DDD.pp
#  West and South are -
sub axu_decode_lat{
   my($tmp)=@_;
   $tmp =~ s/  \.  /30.00/;   
   $tmp =~ s/ \.  /5\.00/;   
   $tmp =~ s/\.  /\.30/;   
   $tmp =~ s/(\.[0-9] ) /${1}5/;   
   $tmp =~ /([0-9][0-9])([0-9][0-9].[0-9][0-9]+)(S|N)/;
   my($deg)=$1 + ($2/60.0);
   $deg = - $deg if $3 eq "S";
   return($deg);
}


# convert degrees (+/-)DDD.pp to text posit (DDDMM.pp[NWES])
sub axu_encode_lat{
   local($degrees)=@_;
   local($sym)="N";
   if ($degrees < 0) {
      $sym="S";
      $degrees=-$degrees;
   }
   local($deg,$min);
   $deg = int ($degrees);
   $min = 60 * ($degrees - $deg);
   return(sprintf("%02d%04.2f%s",$deg,$min,$sym));
}

sub axu_decode_long{
   local($tmp)=@_;
   $tmp =~ s/  \.  /30.00/;   
   $tmp =~ s/ \.  /5\.00/;   
   $tmp =~ s/\.  /\.50/;   
   $tmp =~ s/(\.[0-9]) /${1}5/;   
   $tmp =~ /([0-9][0-9][0-9])([0-9][0-9].[0-9][0-9]+)(E|W)/;
   local($deg)=$1 + ($2/60.0);
   $deg = - $deg if $3 eq "W";
   return($deg);
}
sub axu_encode_long{
   my($degrees)=@_;
   my($sym)="E";
   if ($degrees < 0) {
      $sym="W";
      $degrees=-$degrees;
   }
   my($deg,$min);
   $deg = int ($degrees);
   $min = 60 * ($degrees - $deg);
   return(sprintf("%03d%05.2f%s",$deg,$min,$sym));
}


# calculate range and bearing for two positions

# Statute miles = 1.15078 * Nautical Miles
# Kilometers = 1.852 * Nautical Miles

#Result is in miles.  Adjust $RADIUS to make it km, or whatever.


sub xdist
{
  my ($lat_a, $long_a, $lat_b, $long_b) = @_;
  my ($d);
  $RADIUS = 3963.1;		# radius of the earth in miles (6378 km)

  $d_long = $long_b - $long_a;
  $dist = $RADIUS * &acos ( (sin(&rad($lat_a)) * sin(&rad($lat_b)))
			   + (cos(&rad($lat_a))
			      * cos(&rad($lat_b))
			      * cos(&rad($d_long))
			      )
			   );
}

sub xacos
{
  local ($x) = @_;
  local ($t);

  if ($x == -1.0) {
    # If X is -1, the general formula blows up (zero divided by zero loses),
    # but we know that acos(-1) = pi.
    $t = atan2(1.0, 0.0);
  } else {
    $t = atan2 (sqrt ((1.0 - $x) / (1.0 + $x)), 1.0);
  }
  return ($t + $t);
}

  
sub xrad
{
  local ($deg) = @_;
  $PI = 3.141592653589793;
  return ($deg * $PI / 180.0);
}


sub axu_range()
{  # QUICK routine
  my( $lat1,$lng1,$lat2,$lng2) = @_;
  return &axu_qrange( $lat1, $lng1, $lat2, $lng2);
}

sub axu_bearing()
{  # QUICK routine
  my( $lat1,$lng1,$lat2,$lng2) = @_;
  return 0 unless $lat1;
  return 0 unless $lat2;
  my $angle = atan2($lng2 - $lng1, $lat2 - $lat1);
  $angle += (2*3.14159) if ($angle < 0);
  return ($angle  * 180.0 / 3.14159);
}



# calculate Quick range - using flat geometry
# my ($DistanceMiles) = axu_qrange( $lat1, $lng1, $lat2, $lng2 );
sub axu_qrange()
{
  my( $lat1,$lng1,$lat2,$lng2) = @_;
printf "axu_qrange: %s %s %s %s\n",$lat1,$lng1,$lat2,$lng2 if $debugv;
  return -1 unless $lat1;
  return -1 unless $lat2;
  return -1 if ( ($lat2 < .1) && ($lng2 < .1) );
  my $dist = sqrt( ($lat1-$lat2)*($lat1-$lat2) + ($lng1-$lng2)*($lng1-$lng2));
  return $dist * 60 * 1.15078;
}


sub axu_qbearing()
{
  my( $lat1,$lng1,$lat2,$lng2) = @_;
printf "axu_qbearing: %s %s %s %s\n",$lat1,$lng1,$lat2,$lng2 if $debugv;
  return -1 unless $lat1;
  return -1 unless $lat2;
  return -1 if ( ($lat2 < .1) && ($lng2 < .1) );
  my $angle = atan2($lng2 - $lng1, $lat2 - $lat1);
  $angle += (2*3.14159) if ($angle < 0);
  return ($angle  * 180.0 / 3.14159);
}

# Decode the symbol from a TO string.  GPSxx where xx is MV, LJ, etc
# GPSCnn and GSPEnn
# http://www.dididahdahdidit.com/text/symbols.txt

sub axu_symbol_gps
{
  my( $tostring ) = @_ ;
  $_ = $tostring;
  s/^GPS(..)(.){0,1}/$1/;
  my $overlay = $2;
  my $symbol_found = "?" . $_;
  # GPSLx - symbol is lowercase version of x
  # GPSPx - symbol is uppercase version of x
  # GPSH(S-X)
  # Note:  ord("B") = 66, ord("S") = 83
  SWITCH: {
    # main symbol table
    /B([B-P])/ and $symbol_found = "/" . substr("!\"#$%&'()*+,-./", ord($1)-66, 1) , last SWITCH;
    /P([0-9A-Z])/ and  $symbol_found = "/$1",        last SWITCH ;
    /M([R-X])/ and $symbol_found = "/" . substr(":;<=>?@", ord($1)-82, 1) , last SWITCH;
    /H([S-X])/ and $symbol_found = "/" . substr("\[\\\]\^\_\`", ord($1)-83, 1) , last SWITCH;
    /L([A-Z])/ and  $symbol_found = "/" . lc($1), last SWITCH ;
    /J([1-4])/ and  $symbol_found = "/" . substr("{|}~", ord($1)-ord("{"), 1), last SWITCH ;

    # alternate table entries
    /O([B-P])/ and $symbol_found = "\\" . substr("!\"#$%&'()*+,-./", ord($1)-66, 1) , last SWITCH;
    /A([0-9A-Z])/ and  $symbol_found = "\\$1",        last SWITCH ;
    /N([R-X])/ and $symbol_found = "\\" . substr(":;<=>?@", ord($1)-82, 1) , last SWITCH;
    /D([S-X])/ and $symbol_found = "\\" . substr("\[\\\]\^\_\`", ord($1)-83, 1) , last SWITCH;
    /S([A-Z])/ and  $symbol_found = "\\" . lc($1), last SWITCH ;
    /Q([1-4])/ and  $symbol_found = "/" . substr("{|}~", ord($1)-ord("{"), 1), last SWITCH ;

    # GPS Cnn and Enn entries
    /C(..)/ and  $symbol_found = "/"  . chr(32 + $1),      last SWITCH ;
    /E(..)/ and  $symbol_found = "\\" . chr(32 + $1),      last SWITCH ;
  }
  return $symbol_found . $overlay;
}

# Decode the symbol from SSID 
#  This is obsolete but still sometimes useful - and used
sub axu_symbol_ssid
{
  ($_) = @_ ;
  # remove everthing but the SSID number
  #s/^.*\-(\d{0,2})/$1/;
  s/^[^-]*//;
#  SWITCH: {
#      $ssid_icon =~/-1$/  && do { $aprs_icon = "/a"; last SWITCH;}; # Ambulance
#      $ssid_icon =~/-2$/  && do { $aprs_icon = "/U"; last SWITCH;}; # Bus
#  }
  my $symbol_found = "\\0"; # Dot
  $symbol_found = "/a" if /1/;  # Ambulance
  $symbol_found = "/U" if /2/;  # Bus
  $symbol_found = "/f" if /3/;  # Fire Truck
  $symbol_found = "/b" if /4/;  # Bicycle
  $symbol_found = "/Y" if /5/;  # sail vessel
  $symbol_found = "/X" if /6/;  # Helicopter
  $symbol_found = "/'" if /7/;  # Small Aircraft
  $symbol_found = "/s" if /8/;  # power boat
  $symbol_found = "/>" if /9/;  # car
  $symbol_found = "/<" if /10/; # Motorcycle
  $symbol_found = "/O" if /11/; # Balloon
  $symbol_found = "/j" if /12/; # jeep
  $symbol_found = "/R" if /13/; # RV
  $symbol_found = "/k" if /14/; # Truck
  $symbol_found = "/v" if /15/; # Van
  return $symbol_found;
}


# Given a packet, return the symbol or null if none.
sub axu_symbol
{
  my( $packet ) = @_ ;
  my $symbol_found = "";
  my ($source,$type,$dest,$path,$data) = &axu_decode($packet);
  # trivial rejection of types that have no symbols
  next if $type =~ /_:T>/;
  # MIC-E values - for debugging
  if ( $type =~ m/\'|\`/ ) {
    my($valid, $lat,$lng,$symbol,$status,$course,$speed,$altitude,$message)=axu_decode_MIC_E($dest,$data);
    $symbol_found = $symbol;
  } elsif ( $type =~ m/\@|\!|\=|\;/ ) {
    my($lat,$lng,$symbol,$phg,$course,$speed,$altitude,$text)=axu_decode_position($data);
    $symbol_found = $symbol;
  } elsif ( $type eq "\$" ) {
    # GPS - use TO field or SSID
    $symbol_found = &axu_symbol_gps( $dest );
    # if it was a short GPS, not GPSxx, entry, use the SSID
    if ($symbol_found =~ /^\?/) {
      # try with SSID
      $symbol_found = &axu_symbol_ssid( $source );
      }
    # adjust for overlays
    #$symbol_found = &axu_symbol_overlay($symbol_found);
  }

  return $symbol_found;
}

sub axu_symbol_overlay {
    my ($symbol) = @_;
    # Some of the alternate table symbols allow for an overlay letter
    # or number.   The valid overlays are either 0-9 or A-Z.
    # In some cases the overlay value will be specified in place
    # of the "\" table indicator.  If so, the overlay character
    # should be placed at the end of the list.
    # For example:  a node symbol is broadcast as "N" and "#"
    # The resulting symbol will be "\#N"
    if ($symbol =~ /([A-Z0-9])(.)/ ) {
      $symbol = "\\" . $2 . $1;
    }
    return $symbol;
}


# is packet have a bulletin
# my ($id, $msg ) = &axu_bulletin( $data );
#  if ($id) { # it is a bulletin }

# N7NVP>APW247,K7NWS-10*,WIDE3::BLN2     :generate a few test messages of two or three lines of text.{001

sub axu_bulletin {
  my $data = shift;
  return unless $data;
  if ( $data =~ /:BLN/ ) {
       # a bulletin - $1 == id,  $2 = message
       $data =~ /:(BLN\S+)\s*?:([^{]+)/;
       my $bln_id = $1;
       my $bln_msg = $2;
       return ( $1, $2);
  }
  return ;
}


# Reading a position file doesn't really belong here - but it
# uses lots of axu routines so this is a good place for it for now

sub axu_posfile {
   my ($pos_file_name, $reflat, $reflng, $refname) = @_;

   # set a flag if we want to quick calculate range and bearing
   my $do_calc = "";
   $do_calc = "1" if ( defined $reflat and defined $reflng );
     # print "Doing calculations: $do_calc  lat  $reflat: lng: $reflng\n";

   #my $posdir  = &nwprs_dir("Positions");
   my $posdir  = "/home/nwprs/lib/positions/";
   my $posfile = $posdir . $pos_file_name ;
     # printf "dir: $posdir  full path: $posfile\n";
  
    # we have a potential filename, open it if it's there
    if ( -f $posfile ) {
      open FILE6, $posfile;
      while (<FILE6>) {
        next if ( /^\s*$/ );       # nothing but white space
        next if ( /^\s*[\*\#]/ );  # two possible comment characters
        # add entries to array
        my ( $pos_name, $pos_lat, $pos_lng, $pos_sym, $pos_txt ) = &axu_obj_decode($_);
        # remove white space from position name
        $pos_name =~ s/\s//g;
        # remove new line for text if needed
         $pos_txt =~ s/[\r\n]+$// ;
        # add "x" to the end of the name until it's unique
  #      if ( exists $refname{ $pos_name } ) {
  #printf "posname %s\n",$pos_name;
  #        $pos_name += "x";
  #      }
        $ref_name{ $pos_name } = $pos_txt;
        $ref_lat{ $pos_name }  = $pos_lat;
        $ref_lng{ $pos_name }  = $pos_lng;

        # now calculate the range and bearing to the reference, IF a ref is given
        if ( $do_calc ) {
      #     $ref_range{ $pos_name } = &axu_qrange($reflat,$reflng,$pos_lat, $pos_lng);
      #     $ref_bearing{$pos_name} = &axu_bearing($reflat,$reflng,$pos_lat, $pos_lng);
          ( $ref_range{ $pos_name }, $ref_bearing{ $pos_name } ) =
             &axu_range_and_bearing( $reflat, $reflng, $pos_lat, $pos_lng);
        }
      }
      close FILE6;
    }
  
   # sorted results and put in return array of lists
   my @result;
   foreach $a ( sort { $ref_range{$a} <=> $ref_range{$b} } keys %ref_name ) 
   {
     push @result, [ ( $a, $ref_lat{ $a }, $ref_lng{ $a }, $ref_range { $a }, $ref_bearing{ $a }, $ref_name{ $a } ) ];
   }

  return @result;
}


sub axu_direction_name {
  my ($degrees) = @_;
  # add 11.25 to move the 348.75 to 360 to 360+
  $degrees += 11.25;
  $degrees %= 360;  # get remainder
  return ("N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE", "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW" )[ int( $degrees / 22.5 ) ];
}

# Round Earth Calculation -  http://perlmonks.thepen.com/40354.html
##!/usr/local/bin/perl -w
#
#use strict;
use POSIX qw(atan acos floor);
#
#
#  Use www.mapblast.com, or your coordinates from the MonkMap page if you're already there.
#  Latitude and longtitude are expressed in decimal notation.  North and east values are
#  positive, South and west are negative.
#
#{
#   my @location_of_me    = qw(34.172500 -84.001667);  # latitude=34.172500, longitude=-84.001667 of jcwren
#   my @location_of_vroom = qw(42.763333 -86.110556);  # latitude=42.763333, longitude=-86.110556 of vroom
#
#   print sprintf ("\nMy T-shirt has to travel %0.f miles, on a heading of %.1f degrees, to get from vroom to me\n\n",
#                  range_and_bearing (@location_of_vroom, @location_of_me));
#}
#
#
#  Returns range (in miles), and angle (relative to $latitude1/$longtitude1)
#
#  Uses basic great circle route calculation.  Output compares favorably with www.mapblast.com
#  www.mapquest.com, and Delorme Street Atlas.  Minor deviations since the earth is not a true
#  sphere, but bulges slighty at the equator.  Accurate enough for common distance calculations, 
#  do not use to calculate fuel usage for cruise missiles targetting (0.2% error seems to be 
#  typical worst case, though)
#
#
# my ($range, $bearing) = &axu_range_and_bearing( $latitude1, $longitude1, $latitude2, $longitude2);
#
sub axu_range_and_bearing
{
   my ($latitude1, $longitude1, $latitude2, $longitude2) = @_;

   my $pi = 3.1415926535897;
   my $radToDeg = 180.0 / $pi;
   my $degToRad = $pi / 180.0;
   my $earthRadius = 3958.9;              # Use 3958.9=miles, 6371.0=Km;

   my $distance = 0;
   my $azimuth = 0;
   my $beta = 0;
   my $cosBeta = 0;
   my $cosAzimuth = 0;

   $latitude1  = $latitude1  * $degToRad;
   $longitude1 = $longitude1 * $degToRad;
   $latitude2  = $latitude2  * $degToRad;
   $longitude2 = $longitude2 * $degToRad;

   if (abs ($latitude1) < 90.0)
   {
      $cosBeta = (sin ($latitude1) * sin ($latitude2)) + ((cos ($latitude1) * cos ($latitude2)) * cos ($longitude2 - $longitude1));
   
      if ($cosBeta >= 1.0)
      {
         return (0.0, 0.0);
      }
   
      #
      #  Antipodes  (return miles, 0 degrees)
      #
      if ($cosBeta <= -1.0)
      {
         return (floor ($earthRadius * $pi * 100.0) / 100.0, 0.0);
      }
   
      $beta = acos ($cosBeta);
      $distance = $beta * $earthRadius;
      $cosAzimuth = (sin ($latitude2) - sin ($latitude1) * cos ($beta)) / (cos ($latitude1) * sin ($beta));
   
      if ($cosAzimuth >= 1.0)
      {
         $azimuth = 0.0;
      }
      elsif ($cosAzimuth <= -1.0)
      {
         $azimuth = 180.0;
      }
      else
      {
         $azimuth = acos ($cosAzimuth) * $radToDeg;
      }
   
      if (sin ($longitude2 - $longitude1) < 0.0)
      {
         $azimuth = 360.0 - $azimuth;
      }
   
      return (floor ($distance * 100.0) / 100.0, floor ($azimuth * 100.0) / 100.0);
   }

   #
   #  If P1 Is North Or South Pole, Then Azimuth Is Undefined
   #
   if (sgn ($latitude1) == sgn ($latitude2))
   {
      $distance = $earthRadius * ($pi / 2 - abs ($latitude2));
   }
   else
   {
      $distance = $earthRadius * ($pi / 2 + abs ($latitude2));
   }

   return (floor ($distance * 100.0) / 100.0, 0.0);
}


#
#  Why is there no intrinsic sign function in Perl?
#
sub sgn
{
   return $_[0] == 0 ? 0 : $_[0] < 0 ? -1 : 1;
}


1;
