
# Amateur Experimental PERL WWW libraries.

# For putting Amateur Radio stations, normally but not exclusively APRS
# on cgi generated web pages.

# currently in: ./site_perl/5.6.1/axw.pm
#  sudo cp axw.pm /usr/lib/perl5/site_perl/5.6.1/axw.pm

# public axw routines:
#
#  $n= axw_map_scale { $ScaleName };
#  $ = axw_map_img { $call, $lat, $lng, $cnt_lat, $cnt_lng, $res, $width, $height, more_points );
#  $ = axw_mapblast_scale { $MapScaleNumber }
#  $ = axw_cleancall { callsign };  # remove @ from callsigns
#  $ = axw_url_findu { callsign };
#  $ = axw_url_findu_wx { callsign };
#  $ = axw_url_findu_raw { callsign, age, "display text" };
#  $ = axw_url_findewe { callsign };
#  $ = axw_url_qrz { callsign };
#  $ = axw_url_buck { callsign };
#  $ = axw_url_fcc { callsign };
#  $ = axw_url_heavensabove { $callsign, $lat, $lng };
#  @ = axw_map_scales();
#  $ = axw_url_terraserve {$lat, $lng, $size [0,1,2], $type [1,2]);
#  $ = axw_new_position { $lat, $lng, $lastLat, $lastLng ) ;
#  ($bgcolor ) =  axw_colors 
#  $ =  axw_HourMinute ( minutes [,seconds] )
#  axw_header
#  axw_footer

# Add soon:
# mapquest: http://www.mapquest.com/maps/map.adp?latlongtype=decimal&latitude=47.54800&longitude=-122.80750&zoom=7

# constants
$axw_homepage = "http://wiki.jnos.org/FindEwe";
sub axw_cgibin { return "http://www.jnos.org/cgi-bin" };
$axw_cgibin = &axw_cgibin; 

# changed from .0002 to .0001 - It wasn't considing it
# different from QFC to the P&R to the Post Office
$axw_moved_threshold = 0.0001;


# has the station moved enough to be considered a new position?
#
# axw_new_position
sub axw_new_position {
  my ( $lat, $lng, $lastLat, $lastLng ) = @_;
  my $movement = abs($lat - $lastLat) + abs($lng - $lastLng);
  # don't return result unless movement has been significant
  # return either x + y change or 0 if change is less then threshold value
  return ( $movement < $axw_moved_threshold ) ? 0 : $movement ;
}


# scale speed - axw_speed_scale( )
# Miles Per hour  (1.15, "MPh")
# Nautical Miles per hour or Knots (1.0, "knots")
# Kilometers per hour (1.15 * 1.6, "KPh")
# LightYears (x, "LY" )  186000MPS * 3600 seconds/hr
# mach - compared to speed of sound (x, "mach" )
# fpf - furlongs per fortnight (x, "fpf" )
# fplf - furlongs per lunar fortnight (x, "fplf" )  29.530589 days

# reference:
# http://www.grc.nasa.gov/WWW/K-12/WindTunnel/Activities/knots_vs_mph.html
# conversions:
# http://www.sutron.com/products/reference/conversions/velocity.htm

sub axw_speed_scale {
  my ($scale) = shift;
  return( 1.0, "kts") if ($scale =~ m/kts/i );
  return( 1.0, "knots") if ($scale =~ m/knots/i );
  return( 1.852, "kph") if ($scale =~ m/kph/i );
  return( 1/(186000*3600), "LY") if ($scale =~ m/ly/i );
  return( 1/667.1, "mach") if ($scale =~ m/mach/i );
  return( 3093.293, "fpf") if ($scale =~ m/fpf/i );
  return( (29.530589*(3093.293/14)), "fplf") if ($scale =~ m/fplf/i );
  # default
  return( 1.15077944, "mph") if ($scale =~ m/mph/i );
  return( 1.15077944, "mph");
}


# Generate IMAGE for map blast
# Example:
#http://www.vicinity.com/gif?&CT=47.70183:-122.12300:8000:&IC=47.70183:-122.12300:100:WA7NWP-9&FAM=myblast&W=600&H=350
# now used at Findu: HREF="http://www.mapblast.com/myblast/map.mb?

sub axw_map_img {
  my $call = shift;
  my $lat  = shift;
  my $lng  = shift;
  my $center_lat  = shift;
  my $center_lng  = shift;
  my $res  = shift;
  my $width = shift;
  my $height = shift;
  my @rest = @_;
  # make sure we have good values
  $lat = 0.0 unless $lat;
  $lng = 0.0 unless $lng;
  $res = 0   unless $res;
  $width = 600 unless $width;
  $height = 350 unless $height;
  my $label = 100;
  my $url;
# 60000 = 1 mile, 600000 = 5, 6000000 = 50
# 10 = red cross, 100 = boxed label
# Findu has 3 maps that use:  8000, 400000, 2000000
  $url  = "http://www.vicinity.com/gif?";
#  $url  = "http://www.mapblast.com/myblast/map.mb?";
  $url .= sprintf "CT=%8.5f:%8.5f:%d",$center_lat,$center_lng,$res;
  $url .= sprintf "&IC=%8.5f:%8.5f:%d:%s",$lat,$lng,$label,$call;
  while ( $#rest > 1 ) {
    $call = shift @rest; $lat = shift @rest; $lng = shift @rest;
    $url .= sprintf "&IC=%8.6f:%9.6f:100:%s",$lat,$lng,$call;
  }
  $url .= sprintf "&FAM=myblast&W=%d&H=%d",$width,$height;
  return $url;
}


# Convert our region names to map scale number
sub axw_map_scale {
  my $TheScale = shift;
  my $result = 0;  # default if no match
  my @names = &axw_map_scales;
  for $iName (1..$#names) {
    if ($TheScale eq $names[$iName])
    {
       $result = $iName;
       last;
    }
  }
  return $result;
}
    

# Convert our region number to Mapblast scale

sub axw_mapblast_scale {
  my $TheScale = shift;
  my $map_scale = 40000;  # default
#SCALER: {
#  if ($TheScale eq "NONE")    {$map_scale =        0, next SCALER};
#  if ($TheScale eq "street")  {$map_scale =     5000, next SCALER};
#  if ($TheScale eq "local")   {$map_scale =    40000, next SCALER};
#  if ($TheScale eq "city")    {$map_scale =    80000, next SCALER};
#  if ($TheScale eq "region")  {$map_scale =   200000, next SCALER};
#  if ($TheScale eq "county")  {$map_scale =   400000, next SCALER};
#  if ($TheScale eq "state")   {$map_scale =  6000000, next SCALER};
#  if ($TheScale eq "country") {$map_scale = 60000000, next SCALER};
#  if ($TheScale =~ /\d+/ ) {$map_scale = $TheScale, next SCALER};
#  }
  @Scales = qw/0 5000 40000 80000 200000 400000 6000000 60000000/;
  $map_scale = $Scales[$TheScale] if ( $TheScale >= 0 and $TheScale <= $#Scales);
  return $map_scale;
}


# Return list of possible mapping scales

sub axw_map_scales {
  return qw/NONE street local city region county state country/;
}


# Clean up a callsign for display
sub axw_cleancall{
   my $call = shift;
   # if 3rd party counter:  xyz@abc, return the xyz
   $call =~ s/(.+)@(.*)/$1/;
   # if 3rd party sender:  @abc,  truncate the @ and return abc
   $call =~ s/^\@//;
   return $call;
}



# Generate URL for FindEse Map display

# http://map.findewe.com/wa7nwp-9
sub axw_url_findewe {
   my $call = shift;
   my $url = sprintf "http://map.findewe.com/%s",$call;
   return $url;
}

# Generate URL for original Findu Map display

# http://map.findu.com/wa7nwp-9
sub axw_url_findu {
   my $call = shift;
   $call = &axw_cleancall( $call );
   my $url = sprintf "http://map.findu.com/%s",$call;
   return $url;
}

# Generate URL for Findu Wx by station page

# http://www.findu.com/cgi-bin/wxpage.cgi?call=WW7RA

sub axw_url_findu_wx {
   my $call = shift;
   my $url = "http://www.findu.com/cgi-bin/wxpage.cgi?call=" . $call;
   return $url;
}

#
# Display raw packets at Findu 
#
# http://www.findu.com/cgi.html#raw
# time=1 option, prepends timestamp to each packet
# if age and length are the same, length isn't needed
sub axw_url_findu_raw {
  my $call = shift;
  my $age = shift;
  my $display = shift;
  my $age_length = $age;
  $call = &axw_cleancall( $call );
  my $return_value  = "<a href=\"http://www.findu.com/cgi-bin/raw.cgi?call=$call";
     $return_value .= "&time=1&start=$age";
     $return_value .= "&length=$age_length" if $age ne $age_length;
     $return_value .= "\" >$display</a>";
  return $return_value;
}

# Generate URL for Aprsworld display

sub axw_url_aprsworld {
   my $call = shift;
   my $url = sprintf "http://db.aprsworld.net/datamart/switch.php?call=%s&table=position",$call;
   return $url;
}


# Generate URL for Findham.net

sub axw_url_findham {
   my $call = shift;
   my $url = sprintf "http://www.findham.net/findham.cgi?callsign=%s",$call;
   return $url;
}


# Generate URL for aprsworld map display
#<IMG SRC="http://maps.aprsworld.net/mapserver/map.php?lat=47.69800&lon=-122.12417&label=WA7NWP-9&icon=aprs_pri_98&scale=street">
#<IMG SRC="http://maps.aprsworld.net/mapserver/map.php?lat=47.69800&lon=-122.12417&label=WA7NWP-9&icon=aprs_pri_98&scale=town">
#<IMG SRC="http://maps.aprsworld.net/mapserver/map.php?lat=47.69800&lon=-122.12417&label=WA7NWP-9&icon=aprs_pri_98&scale=regional">
#
#


# Generate URL for QRZ.com display
# http://www.qrz.com/detail/WA7NWP
sub axw_url_qrz {
   my $call = shift;
   $call = &axw_cleancall( $call );
   $call =~ s/-.*//;  # remove SSID at end if necessary
   my $url = sprintf "http://www.qrz.com/detail/%s", $call;
   return $url;
}


# Generate URL for Buckmaster Callsign Lookup
# http://www.qrz.com/detail/WA7NWP
sub axw_url_buck {
   my $call = shift;
   $call =~ s/-.*//;  # remove SSID at end if necessary
   my $url = sprintf "http://www.buckmaster.com/%s", $call;
   return $url;


# <!-- Buckmaster Callsign Lookup Blank START -->
# <form method=POST action=http://hamcall.net/cgi-bin/do_hamcallexe>
#   <table border="0">
#   <tr bgcolor="#99FFFF">
#     <td>
#       <img src="http://hamcall.net/images/b-icon.gif" width="23" height="23" align="left">
#       Type a callsign: <input name=entry>
#       <input type=submit value="Look up callsign">
#     </td>
#   </tr>
#   </table>
# </form>

 

}

# Generate URL for FCC callsign lookup
# http://www.qrz.com/detail/WA7NWP
# HREF="http://gullfoss2.fcc.gov/cgi-bin/ws.exe/genmen/uls/call_res.hts?db_id=19&rows=&callsign=
sub axw_url_fcc {
   my $call = shift;
   $call =~ s/-.*//;  # remove SSID at end if necessary
   my $url = sprintf "http://gullfoss2.fcc.gov/cgi-bin/ws.exe/genmen/uls/call_res.hts?db_id=19&rows=&callsign=%s", $call;
   return $url;
}


# Generate URL for Terraserve images
# Image for WA7NWP-9 from Findu
# http://terraserver.microsoft.com/image.asp?S=10&T=1&X=2828&Y=26415&Z=10&W=1
# $url = &axw_url_terraserve {$lat, $lng [,$map [,$theme [,$scale]]])
# T=theme  1=aerial, 2=topo
# W=map.  0=small, 1=medium, 2=large
# Scaling:  Thanks to Xastir for this info:
#  zoom 10, scale 200, 1 meter resolution
#  zoom 11, scale 400, 2 meter resolution
#  zoom 12  scale 800, 4 meter resolution...
#  zoom 13  scale ?00, 8 meter resolution...
#  zoom 14  scale ?00, 16 meter resolution...
#  zoom 15  scale ?00, 32 meter resolution...
#  zoom 16  scale ?00, 64 meter resolution...
#
sub axw_url_terraserve {
  my $lat = shift;
  my $lng = shift;
  my $map = 1;
     $map = shift if @_;
  my $theme = 1;
     $theme = shift if @_;
  my $scale = 10;
     $scale = shift if @_;
  my $url = sprintf "http://terraserver.microsoft.com/image.asp?";
    $url .= sprintf "Lat=%8.5f&Lon=%9.5f&W=%d&T=%d&S=%d",$lat,$lng,$map,$theme,$scale;

  ### UTM version.
#  my ($zone, $easting, $northing) = axw_deg2utm($lat, $lng);

#  my $scale = 10;
#  $easting /= 200;
#  $northing /= 200;

#     $url .= sprintf "S=%d&T=1&X=%d&Y=%d&Z=%d&W=1", $scale, $easting, $northing, $zone;
  ###
  return $url;
}


# Convert our region number to TerraServer scale

sub axw_terraserve_scale {
  my $TheScale = shift;
  my $map_scale = 10;  # default
  @Scales = qw(10 10 11 12 13 14 15 16);
  $map_scale = $Scales[$TheScale] if ( $TheScale >= 0 and $TheScale <= $#Scales);
  return $map_scale;
}

# Terraserve works best with UTM.
# Here is conversion routine from Experts-Exchange
# http://www.experts-exchange.com/Programming/Q_20176454.html#1
# my ($zone, $easting, $northing) = axw_deg2utm($latitude, $longitude);
# APRSPLUS UTM conversion:  From 47.6986, -122.12517
#    yields UTM of  Zone 10T  Easting 565637  Northing 5283172
#
# My UTM location: S=10&T=1&X=2828&Y=26415&Z=10&W=1
#
# S=scale.  10= 1 meter,  11=2 meters, 12 = 4 meters, 13 = 8 meters
# X=
# Y=
# Z=zone
# T=theme  1=aerial, 2=topo
# W=map.  0=small, 1=medium, 2=large
#
sub axw_deg2utm {
  my ($latitude, $longitude) = @_;
  my $zone = int(($longitude+180)/6)+1;
  my $l = $longitude - $zone*6 +183;
  my $easting = int(1000*cos($latitude*3.1415926535/180)*$l*10000/90+500);
  my $northing = int(1000*10000*$latitude/90);

#printf "lat: $latitude  lng: $longitude -> zone: $zone  easting: $easting  northing: $northing\n";

  return ($zone, $easting, $northing);
}


# Generate URL for Topo maps

# http://www.topozone.com/map.asp?lat=47.69800&lon=-122.11900&s=100&size=s
sub axw_url_topo {
  my $lat = shift;
  my $lng = shift;
  my $size = "s";
  my $url = sprintf "http://www.topozone.com/map.asp?";
    $url .= sprintf "lat=%9.5f&lon=%10.5f&s=100&size=%s", $lat, $lng, $size;
  return $url;
}


#  Generate URL for www.Heavens-Above.com
#  $ = axw_url_heavensabove { $callsign, $lat, $lng };
# Example: http://www.heavens-above.com/amateursats.asp?lat=%2047.491&lng=-122.237&alt=60&loc=WA7NWP-9&TZ=PST

sub axw_url_heavensabove {
  my $name = shift;
  my $lat = shift;
  my $lng = shift;
  my $url = sprintf "http://www.heavens-above.com/amateursats.asp?";
    $url .= sprintf "lat=%7.3f&lng=%8.3f&alt=60&loc=%s&TZ=PST", $lat, $lng, $name;
  return $url;
}


# formatting functions

sub axw_HourMinute {
   my $minutes = shift;
   my $seconds = shift;
   my $hours = 0;
   my $result;
   $minutes += int($seconds /60) if $seconds;
   if ( $minutes > 120 ) {
      $hours = int ( $minutes / 60 );
      $minutes = $minutes % 60;
   }
   $result .= "$hours hours" if ( $hours > 1);
   $result .= "1 hour" if ( $hours == 1);
   $result .= " $minutes minutes" if ( $minutes > 1);
   $result .= " 1 minute" if ( $minutes == 1);
   return $result;
}


# Print common web page color themes urls
# axw_themes ( param, pgm url )

sub axw_themes {
   my $param = shift;
   my $pgm = shift;

   # If a theme exists, remove it
   $param =~ s/theme\=\w+//;

   if ( $pgm ) {
      print "<center>\n";
      print " Colors:";
      print " <a href=\"$pgm\">default</a>\n";
      print " <a href=\"$pgm" . &axw_url_param($param . "&theme=1") ."\">1</a>\n";
      print " <a href=\"$pgm" . &axw_url_param($param . "&theme=nemisis") ."\">NEMISIS</a>\n";
      print " <a href=\"$pgm" . &axw_url_param($param . "&theme=aprsd") ."\">APRSD</a>\n";
      print "</center>\n";
   }
}

# return a good URL parameter string.

# First entry will be a ? - if the string is non-null
# Interior entries will all be &, not ?
# no double entries

sub axw_url_param {
   my $parm = shift;
   return unless $parm;

   # remove first entry if it's a ? or &
   $parm =~ s/^[\?\&]//;

   # make sure it doesn't end with a &
   $parm =~ s/\&$//;

   # force first entry to ?
   return "?" . $parm;
}


# common web page header

sub axw_header {
   my $parms = shift;
      $parms = axw_url_param( $parms );
   print "<a href=\"$axw_cgibin/ewe.cgi$parms\">ALL</a> stations\n";
   print "- <a href=\"$axw_cgibin/ewewx.cgi$parms\">Weather</a>\n";
   print "- <a href=\"$axw_cgibin/ewetlm.cgi$parms\">Telemetry</a>\n";
   print "- <a href=\"$axw_cgibin/ewemsg.cgi$parms\">Messages</a>\n";
}


# common web page footer
sub axw_footer {
   my $url = shift;
   print "<center>\n";
   print "<a href=\"$axw_homepage\">FINDEWE Home</a>\n";
   print "</center>\n";
}


# create hot link where ever http:// or www. is found
# In either case, display the URL, hide the http:

sub axw_url_expand0()
{
  my $reflink = shift;
  # make http://xxx strings a hot link
  $reflink =~ s/http:\/\/(\S+)/<a href=\"http:\/\/$1\">$1<\/a>/; 
  # make " www.xxx" strings a hot link
  $reflink =~ s/([^\/]{0,1})(www\.\S+)/$1<a href=\"http:\/\/$2\">$2<\/a>/;
  return $reflink;
}

# Use all text in argument for name if URL is present
sub axw_url_expand()
{
  my $reflink = shift;
  my $result;
  if ( $reflink =~ /(.*)(http:\/\/\S+)\s(.*)/ )
  {
    # It has a "http://" string.  So the first match becomes the label,
    # and the second match, $2, is the link
    my ($headdata, $link, $taildata) = ($1,$2,$3);
    # at least one real character must be in link
    # strip http off link if it's there
    $link =~ s/^http:\/\///;
    $result =  sprintf "<a href=\"http://%s\">%s</a>",
          $link, $headdata . $taildata ;
  }
  elsif ( $reflink =~ /(.*)(www\.\S+)\s{0,1}(.*)/ )
  {
    # It has a "www." string. 
    # and the second match, $2, is the link
    my ($headdata, $link, $taildata) = ($1,$2,$3);
    # at least one real character must be in link
    $result =  sprintf " <a href=\"http://%s\">%s</a> ",
          $link, $headdata . $taildata;
  }
  else
  {
    # No "www." or "http://" so it's not a hot link
    $result =  $reflink;
  }
  return $result;
}

###
### Email
###

#$In{'email'} = $Email;
#$In{'firstname'} = "boB";

sub email_test {
my $email = <<THE_EMAIL;
From: vodall\@pugetsound.net
To: [[email]]
Subject: My first mailer to [[firstname]] 
 
This is the auto-response email sent when I launched the  
script with a "?$Email" following the script's URL. 
 
Yeah! 

THE_EMAIL

for(keys %In) { $email =~ s/\[\[$_\]\]/$In{$_}/sig; }

$email =~ s/\[\[.*?\]\]//sig;

 &SendEmail("vodall\@pugetsound.net", $Email, "First mailer", $email);
}

# &axw_ValidEmail( $EmailAddress)
# returns 1 if ok, 0 if not

sub axw_ValidEmail
{
	return 0 unless $_[0];
	return 0 if $_[0] =~ /(?:[\.\-\_]{2,})|(?:@[\.\-\_])|(?:[\.\-\_]@)|(?:\A\.)/;
	return 1 if $_[0] =~ /^[\w\.\-\_]+\@\[?[\w\.\-\_]+\.(?:[\w\.\-\_]{2,}|[0-9])\]?$/;
	return 0;
} # sub ValidEmail


# &axw_SendEmail( $from, $to, $subject, $message )
# returns 1 if sent ok.  0 if failure

sub axw_SendEmail
{
  my ( $from, $to, $subject, $message ) = @_;
  return 0 unless &axw_ValidEmail($to);
  my $Mailer = '/usr/sbin/sendmail -t'; 
  my $email;
  $email .= "From: $from\n";
  $email .= "To: $to\n";
  $email .= "Subject: $subject\n";
  $email .= "\n";
  $email .= $message;
  
  open MAIL,"|$Mailer";
  print MAIL $email;
  close MAIL;
  return 1;
}


# Generic webpage routines

sub axw_anchor {
  my($link, $label )=@_;
  # add a "http://" by default unless it already has a protocol
  $link = "http://" . $link unless ( $link =~ m/^\w+\:/ );
  return sprintf( "<A HREF=\"%s\" >%s</A>", $link, $label);
}


# Set colors for web display
#
# ( $bg_color, $font_color, $tbl_hdr_color, $tbl_bg_color, $tbl_special )
#

sub axw_color {
  my $color_theme = shift;
     $color_theme = "default" unless $color_theme;

  ($color_theme eq "1") && return ( "#606060", "#000000", "#FFD700", "#D0D0D0", "#F07070" );
  ($color_theme eq "aprsd") && return ( "#909090", "#000000", "#FFD700", "#D0D0D0", "#F07070" );
  # Nemisis - Star Trek
  # Font Yellow  "#FFFF00", on bright Green "019C00"
  # 66FF66 - "#9CFF64"
  ($color_theme eq "nemisis") && return ( "#005C01", "#9CFF64", "#019C00", "#66FF66", "#005C01" );
  # default
  return ( "#c0c0ff", "#FFD700", "#D0D0D0", "#D0D0F0", "#005C01" ) ;
}

# some special characters are recognized by the http browsers and
# must be converted to their safe counterparts
# > is &gt;
# < is &lt;

sub axw_wwwfix {
  my $s = shift;
  $s =~ s/\>/&gt;/g;
  $s =~ s/\</&lt;/g;
  return $s;
}


#  end of module
1;
